/*
* $Header: /home/dto/octal/RCS/engine.c,v 1.3 2000/07/19 08:00:06 dto Exp dto $
* (C) 2000 David O'Toole $Date: 2000/07/19 08:00:06 $
* 
* Code for OCTAL's signal network engine.  
*
* $Revision: 1.3 $
*
* This software is distributed under the terms of the
* GNU General Public License (GPL). Read the included file
* COPYING for more information. 
*
*/
static const char rcsid[]="$Id: engine.c,v 1.3 2000/07/19 08:00:06 dto Exp dto $";

#include "machine.h"
#include "octal.h"
#include "engine.h"

#define DEFAULT_BLOCK_SIZE (512)
#define DEFAULT_TPB (4)
#define DEFAULT_BPM (120)

/* initialize the engine's internal tables and such */ 

engine* create_engine(void) 
{
  int i, j;
  engine* e;
 
  e = ox_alloc(sizeof(engine));
  if (!e) ox_die("create_engine: couldn't allocate memory");

  /* initialize the machine table */
  for (i=0; i<max_machines; i++) {   
    e->machines[i] = 0;
    for(j=0; j<max_machines; j++)
      e->signals[i][j] = en_nosignal; 
  }  
  e->num_machines = 1; /* we always have master */ 
  e->global_frame = 0; 
  e->tick = 0;
  e->tickpos = 0;
  set_timings(e, DEFAULT_BLOCK_SIZE, DEFAULT_TPB, DEFAULT_BPM); 

  /* now create a dummy master machine */ 
  
  e->machines[master] = (machine*) malloc(sizeof(machine)); 
  e->machines[master]->x = 0.8; 
  e->machines[master]->y = 0.8; 
  strcpy(e->machines[master]->instance_name, "master");

  return e; 
}


/* set_timings calculates tick size and sets up an engine for running */
void set_timings(engine* e, int block_size, int tpb, int bpm)
{
  double temp; 
  
  e->block_size = block_size;
  e->tpb = tpb;
  e->bpm = bpm; 
  /* now calculate the tick size in samples */ 
  temp = (double)(bpm*tpb)/60.0;      /* this is ticks-per-minute / 60 */ 
  temp = (double)OX_SAMPLING_RATE/temp; /* this gives samples-per-tick   */ 
  e->tick_size = (int)temp;            
}


int add_machine(engine* e, machine* m)
{
  int slot; 

  /* now find the first open space in the table */ 
  /* slot starts at one because machines[0] is reserved for "master" right now */ 
  for(slot=1; (e->machines[slot] && slot<max_machines); slot++); 
  
  if (slot == max_machines)  
    ox_die("Machine table full."); 

  /* otherwise, adopt the machine */ 
  e->machines[slot] = m; 
  e->num_machines++;
  return(slot);  
}


void remove_machine(engine* e, int which)
{
  machine* temp; 

  temp = e->machines[which];
  if (temp == 0) return; 
  destroy_machine(temp);
  e->machines[which] = 0;            /* free memory and make a blank space in table */ 
  e->num_machines--; 
}

  
/* returns a pointer to the machine struct */ 
machine* get_machine(engine* e, int m) 
{
  return(e->machines[m]);
}


/* the following function does not yet prevent cycles in the 
 * signal flow. if you do this, it will crash.
 */ 
int connect_machines(engine* e, sig_type t, int source, int sink) {
  if (is_present(e,source) && is_present(e,sink)) 
    e->signals[source][sink] = t; 
}


/* this is just a shortcut */ 
void disconnect_machines(engine* e, int source, int sink) {
  connect(e, en_nosignal, source, sink); 
}


/* check if a machine actually exists */ 
int is_present(engine* e, int which_machine) {
  if (which_machine==master) return 1;         /* don't forget this!! master always exists */ 
  return( e->machines[which_machine] != 0 );
}


/* obsolete */ 

void do_block(engine* e, int m) 
{
  int i; 
  int depcount = 0;
  machine* temp; 

  /* first fill the dependencies */ 
  for (i=1; i<e->num_machines; i++) {
    if (e->signals[i][m] == en_nosignal)  /* skip ones without signals */  
      continue; 
      
    /* ok, so we have a signal coming from i to m */ 
    do_block(e, i);
    }


  /* now mix everything together */ 
  for (i=1; i<e->num_machines; i++) {
    if (e->signals[i][m] == en_nosignal)
      continue; 

    /* if this is the first dep, overwrite the contents of the mix buffer;
     * otherwise, add to the mix buffer.
     */ 
    if (++depcount == 1) {    
      memcpy(e->lmix, e->machines[i]->lout, e->block_size*sizeof(samp));
      memcpy(e->rmix, e->machines[i]->rout, e->block_size*sizeof(samp));
    }
    else {
    add_copy(e->lmix, e->machines[i]->lout, e->block_size);
    add_copy(e->rmix, e->machines[i]->rout, e->block_size);
    }

  }

  /* now scale down the mixed buffer if neccessary */ 
  if (depcount>1) { 
    bad_scale(e->lmix, e->block_size, depcount);
    bad_scale(e->rmix, e->block_size, depcount); 
  }


  /* finally, process the current machine */ 
  if (m!=master) {
    temp = e->machines[m];
    temp->lin = e->lmix;
    temp->rin = e->rmix; 
    temp->type->ox_work(temp, e->block_size);
  }
}


int is_connected(engine* e, int source, int sink)
{
  return (e->signals[source][sink]); 
}




/* some temporary mixing functions */ 

void add_copy(samp* dest, samp* src, int size)
{
  int x;

  for (x=0; x<size; x++)
    dest[x] += src[x];
}


void bad_scale(samp* buf, int size, int numsignals)
{
  int x;
  samp temp = 1.0/(float)numsignals;  

  for (x=0; x<size; x++)
    buf[x] *= temp; 
}



/* End $Source: /home/dto/octal/RCS/engine.c,v $ */
