/*
* $Header: /home/dto/octal/RCS/machine.c,v 1.8 2000/07/19 08:00:06 dto Exp dto $
* (C) 2000 David O'Toole $Date: 2000/07/19 08:00:06 $
* 
* machine loading, instantiation, communication
* and memory allocation 
* 
* $Revision: 1.8 $
*
* This software is distributed under the terms of the
* GNU General Public License (GPL). Read the included file
* COPYING for more information. 
*
*/
static const char rcsid[]="$Id: machine.c,v 1.8 2000/07/19 08:00:06 dto Exp dto $";
#include <dlfcn.h>  /* for UNIX shared-object support */
#include <stdio.h>
#include <stdlib.h> 
#include <dirent.h> /* directory-handling stuff */
#include <sys/types.h>
#include <string.h>

#include "machine.h"
#include "util.h"

package global_oxapi_pkg; 
static machine_type type_registry[max_types];   /* these are built
                                                   from the machine
                                                   directory */
static int num_types = 0;                       /* how many are loaded? */


/* this function installs the neccessary functions into the global package */ 
void init_global_pkg(void) 
{
  static int is_initialized = 0; 
  package* p = &global_oxapi_pkg;
  
  if (is_initialized == 1) /* lazy evaluation */ 
    return; 
  is_initialized = 1; 
  
  p->alloc = pkg_alloc; 
  p->free = ox_free; 
  p->text2note = pkg_text2note;
  p->note2text = pkg_note2text; 
  p->text2freq = pkg_text2freq;
  p->note2freq = pkg_note2freq; 
}


/* looks in the specified directory for shared objects and attempts to
 * load them as OCTAL plugins 
 */
int load_type_registry(const char* machine_dir) {
   
  char *fn;
  char temp[1024];  /* for the concatenated filename */
  int i, ix = 0;     
  DIR *dp;
  struct dirent *ep;

  dp = opendir(machine_dir);
  
  if (dp) {
    while (ep = readdir(dp)) {
      fn = ep->d_name;
      if ( strcmp(fn + (strlen(fn)-3), ".so") != 0 ) 
	continue; 
      
      /* now build the full pathname in temp[] and load the machine */
      strcpy(temp, machine_dir);
      strcat(temp, "/");          /* don't forget the separator slash */ 
      strcat(temp, ep->d_name); 

      if (load_machine_type(&type_registry[ix], temp))
	ix++; 
    }
    num_types = ix; 
  }
  
  else ox_die("%%OCTAL: Couldn't open machine_dir during call to load_type_registry() \n");

  if (ix == 0)
    ox_die("%%OCTAL: No machines found.");

  /* print out some basic info about the machines */
  printf("Successfully loaded type registry from %s with %d machine types. \n",
	 machine_dir, ix);
  
  for (i=0; i<ix; i++) {
    printf("Loaded %s (%s), having %d parameters. \n", type_registry[i].short_name,
	   type_registry[i].long_name, type_registry[i].num_params);
  }
}


/* retrieves an alias for a type, given a short type name */ 
machine_type* get_type(const char* name) {
  int i;

  if (!num_types) ox_die("%%OCTAL: No machine types loaded. \n");
  
  for (i=0; i<num_types; i++) {
    if (strcmp(name, type_registry[i].short_name) == 0)
      return(&type_registry[i]);
    else continue;  
  }
  return(0);  /* wasn't found */ 
}

  

/* accepts a string filename, fills in the structure pointed to by t
   returns 0 on failure, otherwise returns value returned by machine's
   init() */

int load_machine_type(machine_type *t, const char *fn) {
  int ret;
	
  t->so_handle = dlopen(fn, RTLD_NOW); 
  if (!t->so_handle) { 
    fprintf(stderr, "%%OCTAL: %s: dlopen() returned NULL.\n", fn); 
    return(0); 
  }
			
  t->ox_init    = dlsym(t->so_handle, "ox_init");
  t->ox_create  = dlsym(t->so_handle, "ox_create");
  t->ox_destroy = dlsym(t->so_handle, "ox_destroy");
  t->ox_update  = dlsym(t->so_handle, "ox_update");
  t->ox_work    = dlsym(t->so_handle, "ox_work");
  t->ox_desc    = dlsym(t->so_handle, "ox_desc");
  t->ox_track   = dlsym(t->so_handle, "ox_track");
  	
  if (!t->ox_init || !t->ox_create || !t->ox_destroy 
      || !t->ox_work || !t->ox_desc || !t->ox_update || !t->ox_track) {
    printf("%%OCTAL: %s is not a valid OCTAL machine.\n", fn);
    dlclose(t->so_handle);
    return(0);
  }
  ret = t->ox_init(t);	
  return(ret);			
}
	

int unload_machine_type(machine_type *t) {
	
  dlclose(t->so_handle);
	
  t->so_handle  = 0;
  
  t->ox_init    = 0;
  t->ox_work    = 0;
  t->ox_create  = 0;
  t->ox_destroy = 0;
  t->ox_desc    = 0;
  t->ox_update  = 0;
  t->ox_track   = 0;

  return(1);
	
}


machine* create_machine(machine_type *t) 
{
  machine* m;

  m = malloc(sizeof(machine));  
  
  /* install the package */ 
  init_global_pkg(); 
  m->pkg = &global_oxapi_pkg;

  /* call the machine lib */ 
  t->ox_create(m);
  m->type = t;
  m->frame = 0;
  strcpy(m->instance_name, m->type->short_name);  /* default */ 
  m->x = 0.5; m->y = 0.5; 
  
  allocate_outbuf(m, &m->lout, &m->rout); 

  if (!m->rout || !m->lout)
      ox_die("%%OCTAL: Memory allocation failed during create_machine()\n");

  return m; 
}


void destroy_machine(machine *m) {
  
  m->type->ox_destroy(m);
  free_buf(&m->lout, &m->rout);
  m->type = NULL; 
  free(m);
}


void allocate_outbuf(machine* m, samp** left, samp** right) 
{

  if (is_stereo_out(m)) {
    *left = (samp*) ox_alloc(output_buffer_size);
    *right = (samp*) ox_alloc(output_buffer_size);
  }
  else { 
    *left = (samp*) ox_alloc(output_buffer_size);
    *right = *left;
  }
}

     
void free_buf(samp** left, samp** right) 
{
  if (*right == *left) { /* this means we have a mono signal set */ 
    free(*right);
  } 
  else {
    free(*right); 
    free(*left);
  }
    *right = 0;
    *left = 0; 
}


/* these are boolean inspector functions, 
 * for convenience 
 */ 

int is_stereo_out(machine* m) {
  if (m->type->output_channels == stereo) return 1;
  else return 0;
}


int is_stereo_in(machine* m) {
  if (m->type->input_channels == stereo) return 1;
  else return 0;
}


int wants_input(machine* m) {
  if (m->type->input_channels > 0) return 1;
  else return 0;
}


/* memory stuff */ 

void* ox_alloc(size_t s)
{
  void* t; 

  t = malloc(s);
  /* lock the memory if it worked */ 
  return t; 
}

void* pkg_alloc(size_t s)
{
  void* t;

  t = ox_alloc(s);
  /* destroy machine if this fails */ 
  return t; 
}

void ox_free(void* t)
{
  /* munlock(t); */ 
  free(t);
}


/* extras */ 

void set_visual_pos(machine* m, float x, float y) {
  m->x = (m->x < 0 || m->x > 1) ? 0.5 : x;
  m->y = (m->y < 0 || m->y > 1) ? 0.5 : y;
}



/* End $Source: /home/dto/octal/RCS/machine.c,v $*/




















