/*
* $Header: /home/dto/octal/RCS/squaregen.c,v 1.11 2000/07/19 08:00:06 dto Exp dto $
* (C) 2000 David O'Toole $Date: 2000/07/19 08:00:06 $
* 
* Simple Square Wave Generator Machine for GNU OCTAL. Block-oriented output.
* This is intended as a working example of how to write a machine for the OCTAL API.
* You can also use C++ if you prefer; use ox_wrappers (see the Developer's Guide)
* 
* The API is not set in stone yet. Names/methods/etc will probably change a little. 
* Also, the wavetable and accessor objects are designed but not yet written. Coming soon.  
*
* $Revision: 1.11 $
*
* This software is distributed under the terms of the
* GNU General Public License (GPL). Read the included file
* COPYING for more information. 
* 
*/

static const char rcsid[]="$Id: squaregen.c,v 1.11 2000/07/19 08:00:06 dto Exp dto $";

#include "util.h"    /* utility functions for music  */
#include "machine.h" /* machine interface            */

#include <stdlib.h>
#include <string.h>

/* some info about my machine type, to be passed to OCTAL during ox_init() */ 

enum {ix_note = 0, ix_vol=1};  /* indexes of parameters in an event */
enum {my_track = 0 };          /* we only have one track for now    */
enum {my_num_params = 2};

param_spec my_params[] = { 
  {
    note,             /* how to display it in pattern view */ 
    slider,           /* how to control it in control box  */ 
    "note",
    "Which note to play.",
    0,                /* minimum           */
    NOTE_RANGE,       /* maximum           */ 
    0                 /* ignored for note params           */ 
  },

  {
    small,            /* show as two hex digits, 00-FE */
    slider,           /* control type is slider        */ 
    "vol", 
    "Volume: 00-FE",
    0x00, 
    0xFE, 
    0x80,             /* default volume setting        */ 
  }
};

/* this next structure represents an instance of a particular machine. */ 

typedef struct _my_state { 
  freq pitch;    /* calculated from note param whenever it changes */ 
  samp vol;      /* calculated from vol param whenever it changes  */  
  samp phase;    /* current phase of wave                          */  
  samp last_smp; 
  int play;      /* use this to tell when we have noteoff          */ 

} my_state; 



/* initialize the library and tell OCTAL about yourself */
 
int ox_init(machine_type* t) {
  
  t->long_name        = "David's Simple Squarewave Machine"; 
  t->short_name       = "dtosquare"; 
  t->max_tracks       = 1; 
  t->input_channels   = 1; 
  t->output_channels  = 1;   
  t->num_params       = my_num_params; 

  t->param_specs      = my_params; /* give it the address of my array of param structs */ 

  return(1); 
}



/* create a machine by allocating a new my_state structure, and storing its location in m->state.   */
/* this is safe since I know that OCTAL will never touch the value of this pointer; keeping it in   */
/* the machine struct simplifies keeping track of state objects.                                    */ 
void ox_create(machine* m) {
  my_state* s;

  s = m->pkg->alloc(sizeof(my_state));

  s->vol  = 0x80;     
  s->play = 0;           /* don't start until you get a note */ 
  s->last_smp = 0;
  m->state = (void*)s;  
  m->num_tracks = 1;
  }


/* free the memory when you destroy machines */ 

void ox_destroy(machine* m) {
  m->pkg->free(m->state);  
}


void ox_update(machine* m, int track, int param_index, param value) {

  my_state *s = (my_state*)m->state; 
     
  switch(param_index) {
  case ix_note:
    s->pitch = m->pkg->note2freq(value); 
    if (value = noteoff) s->play = 0; 
    else s->play = 1; 
  case ix_vol:
    s->vol = ((float)value) / 255.0;    /* scale from 0x00..0xFE to floating point 0..1 */  
  }
} 


void ox_desc(char* dest, int which_param, param value) {
  float x;
  int percent; 

  switch (which_param) {
  
  case ix_note: 
    sprintf(dest, "Ignored.");
    break; 

  case ix_vol: 
    x = ((float)value) / 255.0;  /* scale from 0x00..0xFE to floating point 0..1 */ 
    percent = (int)(x * 100);
    sprintf(dest, "%d%%", percent); 
    break;

  default: 
    sprintf(dest, "ERROR"); 
  };
}


int ox_work(machine* m, int block_size) {

 my_state* s;  
 samp* mout;              /* mono output for this machine */ 
 freq phaseinc; 
 freq phase;  
 samp vol, tmp; 
 int i;

 s = (my_state*)m->state;         /* grab our state data address                           */ 
 mout = m->lout;                  /* since lout=rout for mono machines, I could use either */ 
 phaseinc = (freq)(s->pitch/44100.0);
 phase = s->phase; 
 vol = s->vol; 

 if (!s->play) return (0); /* nothing to generate */  

 for (i=0; i < block_size; i++) {
    tmp = (phase >= 0.5) ? vol : -vol;
    /* low-pass filter the output */
    tmp = 0.45*tmp + (1.0 - 0.45)*s->last_smp;
    s->last_smp = tmp;
    mout[i] = tmp;
    phase += phaseinc;
    if (phase >= 1.0) phase = 0.0;
  }

 s->phase = phase;   /* put state data back in state struct */  
 return(1);          /* generated successfully */ 
}

void ox_track(machine* m, int change) { 
  /* leave this blank for effects, or if you don't have polyphony */ 
}


/* End $Source: /home/dto/octal/RCS/squaregen.c,v $ */

